part of services;

class DioService {
  DioService._();

  static final _instance = DioService._();
  void Function()? onUnauthenticated;
  void Function(String? errorTitle, String? errorBody, Exception e)?
      onRegularError;

  factory DioService.getInstance() => _instance;

  //--------------------
  final _dio = Dio();
  bool debugMode = false;

  set options(BaseOptions baseOptions) => _dio.options = baseOptions;

  BaseOptions get options => _dio.options;

  void addHeader(Map<String, String> header) =>
      _dio.options.headers.addAll(header);

  void removeHeader(String header) => _dio.options.headers.remove(header);

  void initInterceptors() {
    _dio.interceptors.add(
      InterceptorsWrapper(
        onRequest: (options, handler) {
          if (debugMode) {
            printWrapped('*' * 10);
            if (options.data != null) printWrapped(options.data.toString());
            if (options.queryParameters.isNotEmpty) {
              printWrapped(options.queryParameters.toString());
            }
            printWrapped(options.headers.toString());
            printWrapped('${options.method} ${options.path}');
          }
          if (options.data != null) {
            options.data =
                FormData.fromMap(options.data as Map<String, dynamic>);
          }
          // Do something before request is sent
          return handler.next(options); //continue
        },
        onResponse: (response, handler) {
          if (debugMode) {
            printWrapped(response.data.toString());
            printWrapped(
                '${response.statusCode} ${response.requestOptions.method} ${response.requestOptions.path}');
            printWrapped('*' * 10);
          }
          // Do something with response data
          return handler.next(response); // continue
        },
        onError: (DioError e, handler) {
          if (debugMode) {
            printWrapped(
                'ERROR[${e.response?.statusCode}] => PATH: ${e.requestOptions.path}');
            printWrapped('*' * 10);
          }
          // Do something with response error
          return handler.next(e); //continue
        },
      ),
    );
  }

  Future<Result<T>> get<T>(
    String api, {
    Map<String, dynamic>? queryParameters,
    String? parameter,
    bool showLoading = false,
    bool handleAuth = true,
  }) async {
    try {
      if (showLoading) BotToast.showLoading();
      final response = await _dio.get<T>(
        parameter == null ? api : '$api/$parameter',
        queryParameters: queryParameters,
      );
      return Result(response: response);
    } catch (e) {
      getErrorMessage(e as Exception, handleAuth);
      return Result(error: e);
    } finally {
      if (showLoading) BotToast.closeAllLoading();
    }
  }

  Future<Result<T>> post<T>(
    String api, {
    Map<String, dynamic>? data,
    Map<String, dynamic>? queryParameters,
    String? parameter,
    bool showLoading = false,
    bool handleAuth = true,
  }) async {
    try {
      if (showLoading) BotToast.showLoading();
      final response = await _dio.post<T>(
        parameter == null ? api : '$api/$parameter',
        data: data,
        queryParameters: queryParameters,
      );
      return Result(response: response);
    } catch (e) {
      getErrorMessage(e as Exception, handleAuth);
      return Result(error: e);
    } finally {
      if (showLoading) BotToast.closeAllLoading();
    }
  }

  Future<Result<T>> put<T>(
    String api, {
    Map<String, dynamic>? data,
    Map<String, dynamic>? queryParameters,
    String? parameter,
    bool showLoading = false,
    bool handleAuth = true,
  }) async {
    try {
      data ??= {};
      data['_method'] = 'PUT';
      if (showLoading) BotToast.showLoading();
      final response = await _dio.post<T>(
        parameter == null ? api : '$api/$parameter',
        data: data,
        queryParameters: queryParameters,
      );
      return Result(response: response);
    } catch (e) {
      getErrorMessage(e as Exception, handleAuth);
      return Result(error: e);
    } finally {
      if (showLoading) BotToast.closeAllLoading();
    }
  }

  Future<Result<T>> delete<T>(
    String api, {
    Map<String, dynamic>? data,
    Map<String, dynamic>? queryParameters,
    String? parameter,
    bool showLoading = false,
    bool handleAuth = true,
  }) async {
    try {
      data ??= {};
      data['_method'] = 'DELETE';
      if (showLoading) BotToast.showLoading();
      final response = await _dio.post<T>(
        parameter == null ? api : '$api/$parameter',
        data: data,
        queryParameters: queryParameters,
      );
      return Result(response: response);
    } catch (e) {
      getErrorMessage(e as Exception, handleAuth);
      return Result(error: e);
    } finally {
      if (showLoading) BotToast.closeAllLoading();
    }
  }

  void getErrorMessage(Exception error, bool handleAuth) {
    String? errorTitle;
    String? errorBody;
    if (error is DioError) {
      switch (error.type) {
        case DioErrorType.response:
          final dioError = error;
          if (dioError is DioError) {
            final title = error.response?.data['message'];
            errorTitle = title != null && title.toString().isNotEmpty
                ? title.toString()
                : error.response?.statusMessage;
            errorBody = dioError.response?.data['errors'] != null
                ? getErrorBody(dioError.response?.data['errors'])
                : '';
            if (dioError.response?.statusCode == 401 && handleAuth) {
              onUnauthenticated?.call();
            }
          }
          break;
        case DioErrorType.cancel:
          errorTitle = 'Request to API server was cancelled';
          break;
        case DioErrorType.connectTimeout:
          errorTitle = 'Connection timeout with API server';
          break;
        case DioErrorType.receiveTimeout:
          errorTitle = 'Receive timeout in connection with API server';
          break;
        case DioErrorType.sendTimeout:
          errorTitle = 'Send timeout with server';
          break;
        case DioErrorType.other:
          errorTitle = error.toString();
          break;
      }
    } else {
      errorTitle = 'Unexpected error occurred ${error.toString()}';
    }
    onRegularError?.call(errorTitle.toString(), errorBody, error);

    if (debugMode) {
      printWrapped(errorTitle.toString());
      printWrapped(errorBody.toString());
    }
  }

  String getErrorBody<T>(T error) {
    final result = error;
    if (result is! Map) return '';
    final messageBody = StringBuffer();
    for (final key in result.keys) {
      final value = result[key];
      if (value is List) {
        for (final line in value) {
          if (_dio.options.headers[HttpHeaders.acceptLanguageHeader] == 'ar') {
            messageBody.write('${tr('keys.$key')} : $line\n');
          } else {
            messageBody.write('$line\n');
          }
        }
      }
    }
    return messageBody.toString().trim();
  }
}

class Result<T> {
  final Response<T>? response;
  final Exception? error;

  Result({this.response, this.error});
}
